/******************************************************************************
 * Game of Life application for ATmega168 board with 96 x 60 pixel LCD.
 * The field of cells fills the entire LCD and wraps around.
 *
 * PB2 output: high while sending a new generation to the LCD,
 *             measure frequency to determine speed (generations per second).
 *
 * VERSIONS:
 *   2014-12-15  Arne Rossius
 *    * first version
 *
 ******************************************************************************/



#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "buttons.h"
#include "lcd.h"
#include "tick.h"

#include "images/gol1.h"
#include "images/gol2.h"
#include "images/gol3.h"
#include "images/gol4.h"
#include "images/gol5.h"



#define BUTTON_EXIT 0
#define BUTTON_IMG1 1
#define BUTTON_IMG2 2
#define BUTTON_IMG3 3
#define BUTTON_IMG4 4
#define BUTTON_IMG5 5
#define BUTTON_SPEED 6



volatile uint8_t speed_timeout = 0;
volatile uint8_t speed_display_timeout = 0;
volatile uint8_t help_display_timeout = 0;

// 100 Hz tick function
static void gol_tick(void) {
	if (speed_timeout) speed_timeout--;
	if (speed_display_timeout) speed_display_timeout--;
	if (help_display_timeout) help_display_timeout--;
}



// Game of Life - main function
void gameoflife(void) {
	uint8_t gol[LCD_PAGES][LCD_WIDTH];
	uint8_t buffer_left[LCD_PAGES], buffer_same[LCD_PAGES],
	        buffer_right[LCD_PAGES], first_col[LCD_PAGES];
	uint8_t tick_id, speed, page, page2, mask, mask2, x, newbyte, n;
	
	memcpy_P(gol, img_gol1, sizeof(gol));
	buttons_clear();
	tick_id = tick_add_function(gol_tick);
	help_display_timeout = 200; // 200 ticks @ 100 Hz = 2 seconds
	speed = 0;
	
	do {
	
		// display current generation
		PORTB |= 1<<2; // PB2 = high
		lcd_image(gol[0]);
		PORTB &= ~(1<<2); // PB2 = low
		
		// display OSD messages
		if (help_display_timeout) {
			lcd_text_P(PSTR("1  ) Exit "), 0,  0, NULL);
			lcd_text_P(PSTR("2-6) Image"), 0,  8, NULL);
			lcd_text_P(PSTR("7  ) Speed"), 0, 16, NULL);
		}
		if (speed_display_timeout) {
			lcd_text_P(PSTR("Speed:     Gen/s"), 0, 0, NULL);
			switch (speed) {
				case 0:
					lcd_text_P(PSTR("max"), 7*CHAR_WIDTH, 0, NULL);
					break;
				case 1:
					lcd_text_P(PSTR(" 14"), 7*CHAR_WIDTH, 0, NULL);
					break;
				case 2:
					lcd_text_P(PSTR(" 10"), 7*CHAR_WIDTH, 0, NULL);
					break;
				case 3:
					lcd_text_P(PSTR("  7"), 7*CHAR_WIDTH, 0, NULL);
					break;
				case 4:
					lcd_text_P(PSTR("  5"), 7*CHAR_WIDTH, 0, NULL);
					break;
				case 5:
					lcd_text_P(PSTR("  3"), 7*CHAR_WIDTH, 0, NULL);
					break;
				case 6:
					lcd_text_P(PSTR("  2"), 7*CHAR_WIDTH, 0, NULL);
					break;
				case 7:
					lcd_text_P(PSTR("  1"), 7*CHAR_WIDTH, 0, NULL);
					break;
			}
		}
		
		// initialize buffer and backup of first pixel column
		for (page = 0; page < LCD_PAGES; page++) {
			buffer_same[page] = gol[page][LCD_WIDTH-1];
			buffer_right[page] = gol[page][0];
			first_col[page] = gol[page][0];
		}
		
		for (x = 0; x < LCD_WIDTH; x++) {
			
			// shift buffer to next column
			for (page = 0; page < LCD_PAGES; page++) {
				buffer_left[page] = buffer_same[page];
				buffer_same[page] = buffer_right[page];
				if (x < LCD_WIDTH-1) {
					buffer_right[page] = gol[page][x+1];
				} else {
					buffer_right[page] = first_col[page];
				}
			}
			
			for (page = 0; page < LCD_PAGES; page++) {
			
				newbyte = 0;
				for (mask = 0x80; mask; mask >>= 1) {
				
					// find number of living neighbours
					n = 0;
					// same row neighbours
					if (buffer_left[page] & mask) n++; // left
					if (buffer_right[page] & mask) n++; // right
					// top row neighbours
					if (mask == 0x80) {
						if (page == 0) {
							page2 = LCD_PAGES - 1;
							mask2 = 0x01 << (LCD_PAGES*8 - LCD_HEIGHT);
						} else {
							page2 = page - 1;
							mask2 = 0x01;
						}
					} else {
						page2 = page;
						mask2 = mask << 1;
					}
					if (buffer_left[page2] & mask2) n++; // top left
					if (buffer_same[page2] & mask2) n++; // top
					if (buffer_right[page2] & mask2) n++; // top right
					// bottom row neighbours
					if (page == LCD_PAGES - 1) {
						if (mask == (0x01 << (LCD_PAGES*8 - LCD_HEIGHT))) {
							page2 = 0;
							mask2 = 0x80;
						} else {
							page2 = page;
							mask2 = mask >> 1;
						}
					} else {
						if (mask == 0x01) {
							page2 = page + 1;
							mask2 = 0x80;
						} else {
							page2 = page;
							mask2 = mask >> 1;
						}
					}
					if (buffer_left[page2] & mask2) n++; // bottom left
					if (buffer_same[page2] & mask2) n++; // bottom
					if (buffer_right[page2] & mask2) n++; // bottom right
					
					// apply Game of Life rules
					if (n == 2) {
						// 2 living neighbours: cell retains current state
						newbyte |= mask & buffer_same[page];
					} else if (n == 3) {
						// 3 living neighbours: cell comes to live
						newbyte |= mask;
					} // other number of living neighbours: cell dies
					
				}
				gol[page][x] = newbyte;
				
			} // for (page)
			
		} // for (x)
		
		// delay
		do {
			// poll buttons
			// images: load image repeatedly as long as the button is held down
			if (buttons[BUTTON_IMG1].pressed) {
				memcpy_P(gol, img_gol1, sizeof(gol));
				speed_timeout = 0;
			}
			if (buttons[BUTTON_IMG2].pressed) {
				memcpy_P(gol, img_gol2, sizeof(gol));
				speed_timeout = 0;
			}
			if (buttons[BUTTON_IMG3].pressed) {
				memcpy_P(gol, img_gol3, sizeof(gol));
				speed_timeout = 0;
			}
			if (buttons[BUTTON_IMG4].pressed) {
				memcpy_P(gol, img_gol4, sizeof(gol));
				speed_timeout = 0;
			}
			if (buttons[BUTTON_IMG5].pressed) {
				memcpy_P(gol, img_gol5, sizeof(gol));
				speed_timeout = 0;
			}
			// speed selection: switch speed once per button press
			if (buttons_read(BUTTON_SPEED)) {
				speed++;
				if (speed > 7) speed = 0;
				speed_display_timeout = 100; // 100 ticks @ 100 Hz = 1 second
				speed_timeout = 0;
			}
			
		} while (speed_timeout && (!buttons[BUTTON_EXIT].pressed));
		
		// restart delay
		switch (speed) {
			case 0: speed_timeout =   0; break; // max Gen/s =>    0 ms
			case 1: speed_timeout =   7; break; //  14 Gen/s =>  ~70 ms
			case 2: speed_timeout =  10; break; //  10 Gen/s =>  100 ms
			case 3: speed_timeout =  14; break; //   7 Gen/s => ~140 ms
			case 4: speed_timeout =  20; break; //   5 Gen/s =>  200 ms
			case 5: speed_timeout =  33; break; //   3 Gen/s => ~330 ms
			case 6: speed_timeout =  50; break; //   2 Gen/s =>  500 ms
			case 7: speed_timeout = 100; break; //   1 Gen/s => 1000 ms
		}
		
	} while (!buttons_read(BUTTON_EXIT)); // loop until EXIT button is pressed
	
	tick_remove_function(tick_id);
	
}
