#!/usr/bin/perl -w
# convert image to binary and/or assembler file
# (C) Arne Rossius 2012
#
# 2012-09-09  first version
# 2014-10-15  write header files for C output
# 2014-10-18  added 'const' in C file
# 2014-11-11  added 'header16' option
# 2014-11-15  added 'upside_down' option

$hor         = 0; # horizontal bytes?
$rev         = 1; # set to 1 for MSB-left or MSB-top
$bin         = 0; # enable binary file output
$asm         = 0; # enable assembler file output
$c           = 1; # enable C file output
$header      = 0; # add header (width, height)
$header16    = 0; # use 16-bit values in header instead of 8-bit
$upside_down = 0; # also generate upside down image

use Image::Magick;

if (! @ARGV) {
	die ("Usage: $0 <Filename.png> [<Filename.png> [...]]\n");
}

foreach $file (@ARGV) {
	if (! -f $file) {
		die ("Image file '$file' not found!\n");
	}
	
	# read image file
	$image = Image::Magick->new;
	$image->Read($file);
	$width = $image->Get('width');
	$height = $image->Get('height');
	print "Converting '$file' ($width x $height) ...\n";
	@pixels = $image->GetPixels(map=>'I', width=>$width, height=>$height, normalize=>1);
	
	# open output file(s)
	$filename = $file;
	$filename =~ s/\.[A-Za-z0-9]+$//;
	$filename_ud = "${filename}_180";
	if ($c) {
		open C, ">$filename.c" or die ("Failed to open C file.\n");
		open H, ">$filename.h" or die ("Failed to open H file.\n");
		if ($upside_down) {
			open C2, ">$filename_ud.c" or die ("Failed to open C file.\n");
			open H2, ">$filename_ud.h" or die ("Failed to open H file.\n");
		}
	}
	if ($asm) {
		open ASM, ">$filename.asm" or die ("Failed to open ASM file.\n");
		if ($upside_down) {
			open ASM2, ">$filename_ud.asm" or die ("Failed to open ASM file.\n");
		}
	}
	if ($bin) {
		open BIN, ">$filename.bin" or die ("Failed to open BIN file.\n");
		if ($upside_down) {
			open BIN2, ">$filename_ud.bin" or die ("Failed to open BIN file.\n");
		}
	}
	
	# output header
	if ($c) {
		$var = $filename;
		$var =~ s/[^a-zA-Z0-9]/_/g;
		$hfile = "#ifndef _" . uc($var) . "_H\n" .
			"#define _" . uc($var) . "_H\n" .
			"#include <stdint.h>\n" .
			"#include <avr/pgmspace.h>\n" .
			"#define IMG_" . uc($var) . "_WIDTH $width\n" .
			"#define IMG_" . uc($var) . "_HEIGHT $height\n" .
			"extern const uint8_t PROGMEM img_${var}[];\n" .
			"#endif\n";
		$cfile = "#include <stdint.h>\n" .
			"#include <avr/pgmspace.h>\n" .
			"const uint8_t PROGMEM img_${var}[] = {\n";
		print H $hfile;
		print C $cfile;
		if ($upside_down) {
			print H2 $hfile;
			print C2 $cfile;
		}
	}
	if ($header) {
		if ($c) {
			if ($header16) {
				$hdr = sprintf "\t%02x, %02x, %02x, %02x,\n",
					$width & 0xFF, ($width >> 8) & 0xFF,
					$height & 0xFF, ($height >> 8) & 0xFF;
			} else {
				$hdr = sprintf "\t%d, %d,\n", $width, $height;
			}
			printf C $hdr;
			if ($upside_down) { printf C2 $hdr; }
		}
		if ($asm) {
			if ($header16) {
				print ASM ".dw $width, $height"; # missing \n is intentional
				if ($upside_down) { print ASM ".dw $width, $height"; }
			} else {
				print ASM ".db $width, $height"; # missing \n is intentional
				if ($upside_down) { print ASM ".db $width, $height"; }
			}
		}
		if ($bin) {
			if ($header16) {
				print BIN pack('n', $width), pack('n', $height);
				if ($upside_down) { print BIN pack('n', $width), pack('n', $height); }
			} else {
				print BIN pack('C', $width), pack('C', $height);
				if ($upside_down) { print BIN pack('C', $width), pack('C', $height); }
			}
		}
	}
	
	# convert image
	if ($hor) {
		$xinc = 8;
		$yinc = 1;
	} else {
		$xinc = 1;
		$yinc = 8;
	}
	
	# regular image
	$n = 0;
	for ($y = 0; $y < $height; $y += $yinc) {
		for ($x = 0; $x < $width; $x += $xinc) {
			if ($c) {
				if ($n % 8 == 0) {
					if ($n > 0) {
						print C ",\n\t";
					} else {
						print C "\t";
					}
				} else {
					print C ", ";
				}
			}
			if ($asm) {
				if ($n % 8 == 0) {
					print ASM "\n.db ";
				} else {
					print ASM ", ";
				}
			}
			$data = 0;
			for ($bit = 0; $bit < 8; $bit++) {
				if ($hor) {
					last if ($x + $bit >= $width);
					$cx = $x + $bit;
					$cy = $y;
				} else {
					last if ($y + $bit >= $height);
					$cx = $x;
					$cy = $y + $bit;
				}
				if ($pixels[$cy * $width + $cx] < .5) {
					if ($rev) {
						$data |= 1<<(7-$bit);
					} else {
						$data |= 1<<$bit;
					}
				}
			}
			if ($c) {
				printf C "0x%02x", $data;
			}
			if ($asm) {
				printf ASM "0x%02x", $data;
			}
			if ($bin) {
				print BIN pack('C', $data);
			}
			$n++;
		}
	}
	# close files
	if ($c) {
		close H;
		print C "\n};\n";
		close C;
	}
	if ($asm) {
		print ASM "\n";
		close ASM;
	}
	if ($bin) {
		close BIN;
	}
	
	# upside down image
	if ($upside_down) {
		$n = 0;
		for ($y = $height - 1; $y >= 0; $y -= $yinc) {
			for ($x = ($width - 1); $x >= 0; $x -= $xinc) {
				if ($c) {
					if ($n % 8 == 0) {
						if ($n > 0) {
							print C2 ",\n\t";
						} else {
							print C2 "\t";
						}
					} else {
						print C2 ", ";
					}
				}
				if ($asm) {
					if ($n % 8 == 0) {
						print ASM2 "\n.db ";
					} else {
						print ASM2 ", ";
					}
				}
				$data = 0;
				for ($bit = 0; $bit < 8; $bit++) {
					if ($hor) {
						last if ($x - $bit < 0);
						$cx = $x - $bit;
						$cy = $y;
					} else {
						last if ($y - $bit < 0);
						$cx = $x;
						$cy = $y - $bit;
					}
					if ($pixels[$cy * $width + $cx] < .5) {
						if ($rev) {
							$data |= 1<<(7-$bit);
						} else {
							$data |= 1<<$bit;
						}
					}
				}
				if ($c) {
					printf C2 "0x%02x", $data;
				}
				if ($asm) {
					printf ASM2 "0x%02x", $data;
				}
				if ($bin) {
					print BIN2 pack('C', $data);
				}
				$n++;
			}
		}
		# close files
		if ($c) {
			close H2;
			print C2 "\n};\n";
			close C2;
		}
		if ($asm) {
			print ASM2 "\n";
			close ASM2;
		}
		if ($bin) {
			close BIN2;
		}
	}
}
