/******************************************************************************
 * LCD Driver for PCF8548 LCD, for Pollin VG-090651 LCD (96 x 60 pixels)
 * Assuming: LCD V_CC2/3 = 3.3 V, external I2C bus pull-up resistors = 4.7k,
 *           using hardware I2C
 *
 * LCD memory:
 *  - 8 "pages" of 96 bytes each, each page is 8 pixels high and 96 pixels wide
 *  - The first byte in page 0 is the top-left pixel and 7 pixels below
 *  - The MSB of each byte is the top pixel row of the page, the LSB is bottom
 *  - In page 7, only the four MSBs (bits 7 to 4) of each byte are used
 *
 * VERSIONS:
 *   2014-12-14  Arne Rossius
 *    * first version
 *
 ******************************************************************************/

#include <stddef.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "lcd.h"



// Send I2C start condition
static void lcd_i2c_start(void) {
	TWCR = 1<<TWINT | 1<<TWSTA | 1<<TWEN;
}

// Send I2C stop condition
static void lcd_i2c_stop(void) {
	while ((TWCR & (1<<TWINT)) == 0); // wait until ongoing transfer completes
	TWCR = 1<<TWINT | 1<<TWSTO | 1<<TWEN;
}

// Send I2C data byte (ACK/NACK is ignored)
static void lcd_i2c_send(uint8_t data) {
	while ((TWCR & (1<<TWINT)) == 0); // wait until ongoing transfer completes
	TWDR = data;
	TWCR = 1<<TWINT | 1<<TWEN;
}



// Start a data transfer to the LCD
static void lcd_start_data(uint8_t page, uint8_t x) {
	lcd_i2c_start();
	lcd_i2c_send(0x78); // I2C address
	lcd_i2c_send(0x80); // Control byte: command + new control byte follows
	lcd_i2c_send(0x80 + x); // Set X address = 0
	lcd_i2c_send(0x80); // Control byte: command + new control byte follows
	lcd_i2c_send(0x40 + page); // Set Y address = page
	lcd_i2c_send(0x40); // Control byte: data follows
}



// Initialize LCD
void lcd_init(void) {
	
	// init TWI module for I2C transfers
	// transmission speed: F_TWI = F_CPU / (16 + 2 * TWBR * 4^TWPS)
#if (F_CPU == 18432000)
	TWBR = 15; // 18.432 MHz / (16 + 2 * 15 * 4^0) = ~400.7 kHz
	TWSR = 0; // set TWPS = 0
#else
	#error "No TWI configuration for selected F_CPU"
#endif
	TWCR = 1<<TWEN; // enable TWI module
	
	// init Timer 1 for backlight PWM: Fast PWM mode, TOP = ICR1
	ICR1 = 10000; // TOP value
	OCR1A = 10000; // backlight = 0% (off)
	TCCR1A = 1<<COM1A1 | 1<<COM1A0 | 1<<WGM11; // inverting PWM mode
	TCCR1B = 1<<WGM13 | 1<<WGM12 | 0x01; // start timer, Clk/1
	
	// Reset LCD
	_delay_ms(10);
	LCD_PORT &= ~(1<<LCD_RESET); // assert Reset
	_delay_us(10);
	LCD_PORT |= 1<<LCD_RESET; // de-assert Reset
	_delay_us(10);
	
	// send init command sequence to LCD
	lcd_i2c_start();
	lcd_i2c_send(0x78); // I2C address
	lcd_i2c_send(0x00); // Control byte: commands follow
	lcd_i2c_send(0x21); // Function Set: Mirroring off, Power on, Increment x,
	                    //               Select extended instruction set (H=1)
	lcd_i2c_send(0x04); // Temperature Coefficient: 0
	lcd_i2c_send(0x08); // Display Config: Top = MSB, Mirroring off
	lcd_i2c_send(0x12); // Bias System: Bias = 1/9 (recommended for 1/65 Mux)
	lcd_i2c_send(0x20); // Function Set: Mirroring off, Power on, Increment x,
	                    //               Select standard instruction set (H=0)
	lcd_i2c_send(0x05); // Set V_LCD Range: upper range (PRS=1)
	lcd_i2c_send(0x11); // Set HV Gen Stages: 3x multiplication
	lcd_i2c_send(0x0C); // Display Control: LCD on, Inverted Video off
	lcd_i2c_send(0x40 + 0); // Set Y Address = 0
	lcd_i2c_send(0x80 + 0); // Set X Address = 0
	lcd_i2c_stop();
	lcd_contrast(50); // Set V_OP = 8.25 V
}

// Set LCD contrast
void lcd_contrast(uint8_t contrast) {
	// with PRS = 1: V_OP = (6.75 + contrast * 0.03) V
	if (contrast > 75) contrast = 75; // limit V_OP to 9 V
	lcd_i2c_start();
	lcd_i2c_send(0x78);
	lcd_i2c_send(0x00); // Control byte: commands follow
	lcd_i2c_send(0x21); // Function Set: Select extended instruction set (H=1)
	lcd_i2c_send(0x80 + contrast); // Set V_OP
	lcd_i2c_send(0x20); // Function Set: Select standard instruction set (H=0)
	lcd_i2c_stop();
}

// Set LCD backlight brightness (0 to 100)
void lcd_backlight(uint8_t brightness) {
	OCR1A = 10000 - brightness * brightness; // Gamma = 2
}



// Clear entire LCD
void lcd_clear(void) {
	uint8_t page, x;
	for (page = 0; page < LCD_PAGES; page++) {
		lcd_start_data(page, 0);
		for (x = 0; x < LCD_WIDTH; x++) {
			lcd_i2c_send(0x00);
		}
		lcd_i2c_stop();
	}
}



// Display full-size image from RAM
void lcd_image(const uint8_t *image) {
	uint8_t page, x;
	for (page = 0; page < LCD_PAGES; page++) {
		lcd_start_data(page, 0);
		for (x = 0; x < LCD_WIDTH; x++) {
			lcd_i2c_send(*image++);
		}
		lcd_i2c_stop();
	}
}

// Display full-size image from flash memory.
void lcd_image_P(const uint8_t *image) {
	uint8_t page, x, data;
	for (page = 0; page < LCD_PAGES; page++) {
		lcd_start_data(page, 0);
		for (x = 0; x < LCD_WIDTH; x++) {
			data = pgm_read_byte(image++);
			lcd_i2c_send(data);
		}
		lcd_i2c_stop();
	}
}

// Display text from RAM, either directly on the LCD (fb == NULL) or into a
// framebuffer in RAM which is sent to the LCD later. If the text is sent to
// the LCD directly and y % 8 != 0, some space above and below the text will
// be overwritten with white pixels.
void lcd_text(const char *text, uint8_t x, uint8_t y,
              uint8_t fb[LCD_PAGES][LCD_WIDTH]) {
	const char *text2;
	const uint8_t *charset_ptr;
	char c;
	uint8_t i, x2, page, offset, mask, data;
	
	text2 = text;
	x2 = x;
	page = y / 8;
	offset = y % 8;
	
	if (fb == NULL) lcd_start_data(page, x);
	mask = 0xFF << (8 - offset);
	while ((c = *text++)) {
		charset_ptr = charset[(uint8_t)c];
		for (i = 0; i < CHAR_WIDTH; i++) {
			data = pgm_read_byte(charset_ptr++);
			if (offset) data >>= offset;
			if (fb == NULL) {
				lcd_i2c_send(data);
			} else {
				if (offset) data |= fb[page][x] & mask;
				fb[page][x++] = data;
			}
		}
	}
	if (fb == NULL) lcd_i2c_stop();
	
	if (offset) {
		page++;
		if (fb == NULL) lcd_start_data(page, x2);
		mask = 0xFF >> offset;
		while ((c = *text2++)) {
			charset_ptr = charset[(uint8_t)c];
			for (i = 0; i < CHAR_WIDTH; i++) {
				data = pgm_read_byte(charset_ptr++);
				if (offset) data <<= (8 - offset);
				if (fb == NULL) {
					lcd_i2c_send(data);
				} else {
					data |= fb[page][x2] & mask;
					fb[page][x2++] = data;
				}
			}
		}
		if (fb == NULL) lcd_i2c_stop();
	}
}

// Display text from flash memory.
// For more information, see lcd_text().
void lcd_text_P(const char *text, uint8_t x, uint8_t y,
                uint8_t fb[LCD_PAGES][LCD_WIDTH]) {
	char ramtext[LCD_WIDTH / CHAR_WIDTH + 1];
	strncpy_P(ramtext, text, sizeof(ramtext));
	ramtext[sizeof(ramtext)-1] = 0;
	lcd_text(ramtext, x, y, fb);
}

// Display a decimal value (specific number of digits, leading zeros blanked).
// For more information, see lcd_text().
void lcd_dec(uint16_t val, uint8_t digits, uint8_t x, uint8_t y,
             uint8_t fb[LCD_PAGES][LCD_WIDTH]) {
	char buf[6];
	buf[0] = buf[1] = buf[2] = buf[3] = buf[4] = '0';
	buf[5] = 0; // string terminator
	while (val >= 10000) { buf[0]++; val -= 10000; }
	while (val >=  1000) { buf[1]++; val -=  1000; }
	while (val >=   100) { buf[2]++; val -=   100; }
	while (val >=    10) { buf[3]++; val -=    10; }
	buf[4] += val;
	if (buf[0] == '0') {
		buf[0] = ' ';
		if (buf[1] == '0') {
			buf[1] = ' ';
			if (buf[2] == '0') {
				buf[2] = ' ';
				if (buf[3] == '0') buf[3] = ' ';
			}
		}
	}
	if (digits > 5) digits = 5;
	lcd_text(&buf[5-digits], x, y, fb);
}
