/******************************************************************************
 * Example Program for ATmega168 board with 96 x 60 pixel LCD
 * ATmega168 @ 18.432 MHz (external crystal)
 *
 * PORTS:
 *   PB0   = Push-button #7 (to ground, internal pull-up)
 *   PB1   = LCD backlight enable [OC1A]
 *   PB2   = Debugging/Measurement output
 *   PB3-5 = [AVR ISP]
 *   PB6-7 = [XTAL]
 *   PC0-2 = [unused]
 *   PC3   = LCD Reset (active low)
 *   PC4   = I2C SDA (external 4.7k pull-up resistor) [SDA]
 *   PC5   = I2C SCL (external 4.7k pull-up resistor) [SCL]
 *   PD0   = UART Rx (internal pull-up) [RXD]
 *   PD1   = UART Tx [TXD]
 *   PD2-7 = Push-buttons #1 to #6 (to ground, internal pull-up)
 *
 * VERSIONS:
 *   2014-12-15  Arne Rossius
 *    * first version
 *
 ******************************************************************************/



#include <stddef.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "buttons.h"
#include "gameoflife.h"
#include "lcd.h"
#include "mcuf.h"
#include "tick.h"

#include "images/geese.h"



void __attribute__((OS_main)) main(void) {
	uint8_t backlight = 0, backlight_new;
	char buf[3];

	// init ports (unused pins: input with pull-up)
	DDRB = 0x06;
	PORTB = 0xF9;
	DDRC = 0x08;
	PORTC = 0x0F;
	DDRD = 0x02;
	PORTD = 0xFF;
	
	// init 100 Hz tick system
	tick_init();
	
	// init buttons
	buttons_init();
	
	// init LCD
	lcd_init();
	lcd_clear();
	
	// fade-in LCD backlight
	do {
		_delay_ms(2);
		lcd_backlight(++backlight);
	} while (backlight < 100);
	
	sei(); // enable interrupts
	
	// main loop
	while (1) {
	
		// display main menu
		lcd_clear();
		lcd_text_P(PSTR("Welcome \x02"), (LCD_WIDTH - 9*CHAR_WIDTH) / 2,
		           0, NULL); // page 0 (y = 0~7)
		           // CP437 0x02 = Unicode 0x263B = Black Smiling Face
		lcd_text_P(PSTR("\xC4\xC4\xC4\xC4\xC4\xC4\xC4\xC4\xC4\xC4"),
		           (LCD_WIDTH - 10*CHAR_WIDTH) / 2, 8, NULL); // page 1
		           // CP437 0xC4 = Unicode 0x2500 = Box Drawings Light Horiz.
		lcd_text_P(PSTR("1) Game of Life"), 0, 16, NULL); // page 2
		lcd_text_P(PSTR("2) MCUF Display"), 0, 28, NULL); // page 3-4
		lcd_text_P(PSTR("3) Geese"), 0, 40, NULL); // page 5
		lcd_text_P(PSTR("4) Backlight"), 0, 52, NULL); // page 6-7
		
		// display current backlight setting
		if (backlight == 100) {
			buf[0] = '1';
			buf[1] = '0';
		} else {
			buf[0] = ' ';
			buf[1] = '0' + backlight / 10;
		}
		buf[2] = 0;
		lcd_text(buf, 14*CHAR_WIDTH, 52, NULL);
		
		// wait for button press
		buttons_clear();
		while (1) {
			
			// button 1: Game of Life
			if (buttons_read(0)) {
				gameoflife();
				break;
			}
			
			// button 2: MCUF Stream Display
			if (buttons_read(1)) {
				mcuf_display();
				break;
			}

			// button 2: Image of Geese
			if (buttons_read(2)) {
				// no lcd_clear() necessary because image overwrites entire LCD
				lcd_image_P(img_geese);
				while (!buttons_read(0) && !buttons_read(1) &&
				       !buttons_read(2) && !buttons_read(3) &&
				       !buttons_read(4) && !buttons_read(5) &&
				       !buttons_read(6));
				break;
			}
			
			// button 3: Backlight
			if (buttons_read(3)) {
				// increment backlight brightness in steps of 10%
				if (backlight < 100) {
					// fade to next step
					backlight_new = backlight + 10;
					do {
						_delay_ms(10);
						lcd_backlight(++backlight);
					} while (backlight < backlight_new);
				} else {
					// already at highest setting => fade to zero
					do {
						_delay_ms(1);
						lcd_backlight(--backlight);
					} while (backlight);
				}
				break;
			}
			
		}
	
	} // end of main loop
	
} // end of main()

