/******************************************************************************
 * MCUF (MicroController Unit Frame) Stream Display application for
 * ATmega168 board with 96 x 60 pixel LCD.
 *
 * Supports streams with 115200 Baud, BlinkenLights resolution (18 x 8 pixels),
 * 1 channel (monochrome), maxval = { 1, 3, 7, 15, 31, 63, 127, 255 }.
 *
 * PB2 output: high while sending a new frame to the LCD.
 *
 * VERSIONS:
 *   2014-12-15  Arne Rossius
 *    * first version
 *
 ******************************************************************************/



#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

#include "buttons.h"
#include "lcd.h"
#include "tick.h"
#include "uart.h"



#define MCUF_FRAME_TIMEOUT 10

#define BUTTON_EXIT 0



const uint8_t PROGMEM mcuf_header[] = {
	0x23, 0x54, 0x26, 0x66, // magic
	0x00, 0x08, 0x00, 0x12, // height = 8, width = 18
	0x00, 0x01, 0x00, 0xFF  // channels = 1, maxval = 0xFF
};



volatile uint8_t frame_timeout = 0;

// 100 Hz tick function
static void mcuf_tick(void) {
	if (frame_timeout) frame_timeout--;
}



// MCUF Stream Display - main function
void mcuf_display(void) {
	uint8_t lcd[LCD_PAGES][LCD_WIDTH];
	uint8_t tick_id, maxval, width = 0, height = 0, channels = 0, threshold = 0;
	uint8_t data, mcuf_pos = 0, page = 0, x = 0;
	uint16_t mask = 0x0000, pixels, frames = 0;
	enum {
		FMT_UNKNOWN,
		FMT_BL,
		FMT_ARCADE
	} format = FMT_UNKNOWN;
	
	uart_init(BAUD_115200);
	tick_id = tick_add_function(mcuf_tick);
	buttons_clear();
	
	// display frame and info
	memset(lcd, 0x00, sizeof(lcd));
	lcd_text_P(PSTR("Waiting..."), 0, 0, lcd);
	lcd_text_P(PSTR("1) Exit"), 0, 16, lcd);
	lcd_text_P(PSTR("W"), LCD_WIDTH - 3*CHAR_WIDTH, 0, lcd);
	lcd_text_P(PSTR("H"), LCD_WIDTH - 3*CHAR_WIDTH, 8, lcd);
	lcd_text_P(PSTR("Ch"), LCD_WIDTH - 3*CHAR_WIDTH, 16, lcd);
	lcd_text_P(PSTR("Max"), LCD_WIDTH - 3*CHAR_WIDTH, 30, lcd);
	lcd_image(lcd[0]);
	
	// main loop
	do {
	
		// check for data from UART
		if (uart_data_available()) {
			data = uart_getc();
			frame_timeout = MCUF_FRAME_TIMEOUT;
			
			if (mcuf_pos < 12) {
				// receive header
				switch (mcuf_pos) {
					default:
						if (data == pgm_read_byte(&mcuf_header[mcuf_pos])) {
							// byte is correct
							mcuf_pos++;
						} else {
							// byte is not correct
							mcuf_pos = 0;
						}
						break;
					case 5: // height
						height = data;
						mcuf_pos++;
						break;
					case 7: // width
						width = data;
						mcuf_pos++;
						break;
					case 9: // channels
						channels = data;
						mcuf_pos++;
						break;
					case 11: // maxval
						maxval = data;
						threshold = data / 2;
						mcuf_pos++;
						// display info
						if (width < 100) {
							lcd_dec(width, 2, LCD_WIDTH - 2*CHAR_WIDTH, 0,
							        lcd);
						} else {
							lcd_text_P(PSTR("##"), LCD_WIDTH - 2*CHAR_WIDTH, 0,
							           lcd);
						}
						if (height < 100) {
							lcd_dec(height, 2, LCD_WIDTH - 2*CHAR_WIDTH, 8,
							        lcd);
						} else {
							lcd_text_P(PSTR("##"), LCD_WIDTH - 2*CHAR_WIDTH, 8,
							           lcd);
						}
						if (channels < 10) {
							lcd_dec(channels, 1, LCD_WIDTH - 1*CHAR_WIDTH, 16,
							        lcd);
						} else {
							lcd_text_P(PSTR("#"), LCD_WIDTH - 1*CHAR_WIDTH, 16,
							           lcd);
						}
						lcd_dec(maxval, 3, LCD_WIDTH - 3*CHAR_WIDTH, 38, lcd);
						// prepare for data reception
						pixels = width * height * channels;
						if ((width == 18) && (height == 8) &&
						    (channels == 1)) {
							// BlinkenLights 144px format
							for (x = 1; x < 2 + 18*4; x++) {
								lcd[0][x] = 0x80; // top line (y=0)
								lcd[7][x] = 0x20; // bottom line (y=58)
							}
							for (page = 0; page < 8; page++) {
								if (page == 7) mask = 0xE0; else mask = 0xFF;
								lcd[page][0] |= mask; // left
								lcd[page][2 + 18*4] |= mask; // right
							}
							page = 0;
							x = 2;
							mask = 0x003F;
							format = FMT_BL;
						} else if ((width == 26) && (height == 20) &&
						           (channels == 1)) {
							// ARCADE 520px format
							for (x = 23 + 1; x < 23 + 3 + 26; x++) {
								lcd[1][x] = 0x80; // top line (y=8)
								lcd[6][x] = 0x10; // bottom line (y=51)
							}
							for (page = 1; page < 7; page++) {
								if (page == 6) mask = 0xF0; else mask = 0xFF;
								lcd[page][23] |= mask; // left
								lcd[page][23 + 3 + 26] |= mask; // right
							}
							page = 1;
							x = 23 + 2;
							mask = 0x0030;
							format = FMT_ARCADE;
						} else {
							// unsupported format
							format = FMT_UNKNOWN;
							// display message
							lcd_text_P(PSTR("Format not"), 0, 0, lcd);
							lcd_text_P(PSTR("supported!"), 0, 10, lcd);
						}
						break;
				}
			
			} else { // if (mcuf_pos < 12)
				
				// receive data
				switch (format) {
				
					case FMT_BL:
						// display BlinkenLights pixel (size 3x6 + border)
						if (data > threshold) {
							if (page >= 3) {
								lcd[page-1][x]   |= mask >> 8;
								lcd[page-1][x+1] |= mask >> 8;
								lcd[page-1][x+2] |= mask >> 8;
							}
							lcd[page][x]   |= mask & 0xFF;
							lcd[page][x+1] |= mask & 0xFF;
							lcd[page][x+2] |= mask & 0xFF;
						}
						// increment pixel position
						x += 4;
						if (x >= 2 + 18*4) {
							x = 2;
							page++;
							mask <<= 1;
						}
						break;
					
					case FMT_ARCADE:
						// display ARCADE pixel (size 1x2, no border)
						if (data > threshold) {
							lcd[page][x] |= mask;
						}
						// increment pixel position
						if (++x >= 23 + 2 + 26) {
							x = 23 + 2;
							mask >>= 2;
							if (!mask) {
								page++;
								mask = 0xC0;
							}
						}
						break;
						
					case FMT_UNKNOWN:
						// do nothing
						break;
					
				} // switch (format)
				
				if (--pixels == 0) {
					// frame complete
					// display frame counter
					frames++;
					if (frames >= 1000) frames = 0;
					lcd_dec(frames, 3, LCD_WIDTH - 3*CHAR_WIDTH, 52, lcd);
					// display new frame
					PORTB |= 1<<2;
					lcd_image(lcd[0]);
					PORTB &= ~(1<<2);
					// clear RAM buffer for next frame
					for (page = 0; page < LCD_PAGES; page++) {
						for (x = 0; x < LCD_WIDTH - 3*CHAR_WIDTH; x++) {
							lcd[page][x] = 0x00;
						}
					}
					// reset reception position
					mcuf_pos = 0;
				}
				
				
			} // if (mcuf_pos < 12)
		}
		
		// check for timeout
		if (frame_timeout == 0) mcuf_pos = 0;
	
	// loop until EXIT button is pressed
	} while (!buttons_read(BUTTON_EXIT));
	
	tick_remove_function(tick_id);
	uart_disable();
	
}
