/******************************************************************************
 * 100 Hz tick function calling.
 *
 * This file contains a 100 Hz timer interrupt. Other parts of the
 * program can add functions which should be called from this interrupt.
 *
 * VERSIONS:
 *   2014-12-14  Arne Rossius
 *    * first version
 *
 ******************************************************************************/



#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "tick.h"



void (* volatile funcs[MAX_TICK_FUNCTIONS])(void);



// Timer 0 output compare A interrupt (CTC overflow) - 100 Hz
ISR(TIMER0_COMPA_vect) {
	uint8_t i;
	for (i = 0; i < MAX_TICK_FUNCTIONS; i++) {
		if (funcs[i] != NULL) (*funcs[i])();
	}
}



// init tick system
void tick_init(void) {
	
	// init variables
	memset((void*)funcs, 0x00, sizeof(funcs));
	
	// init Timer 0 for 100 Hz tick interrupt: 18.432 MHz / 1024 / 180 = 100 Hz
	OCR0A = 180 - 1; // TOP value
	TIMSK0 = 1<<OCIE0A; // enable output compare A interrupt (CTC overflow)
	TCCR0A = 1<<WGM01; // CTC mode, TOP = OCR0A
	TCCR0B = 0x05; // start timer, Clk/1024
}



// add a function to be executed with 100 Hz,
// return function ID (needed for removal) or 0 for error
uint8_t tick_add_function(void(*func)(void)) {
	uint8_t i;
	for (i = 0; i < MAX_TICK_FUNCTIONS; i++) {
		if (funcs[i] == NULL) {
			cli();
			funcs[i] = func;
			sei();
			return i + 1;
		}
	}
	return 0;
}

// remove a function (no longer execute it with 100 Hz)
void tick_remove_function(uint8_t func_id) {
	if ((func_id == 0) || (func_id > MAX_TICK_FUNCTIONS)) return;
	func_id--;
	cli();
	funcs[func_id] = NULL;
	sei();
}
